/**
 * @file
 *
 * Ce fichier est le coeur de pacman. Il contient toute
 * la logique du jeu
 *
 * @author Sylvain MARECHAL
 * @date $Id$
 */
#ifdef _WIN32
#pragma warning(disable:4996) /*Suppression du warning ""'' was declared deprecated""*/
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "pac_core.h"
#include "pac_err.h"
#include "pac_view.h"
#include "pac_file.h"

/**
 * Structure memorisant 2 positions sucessives d'une 
 * creature afin que le deplacement de la creature
 * semble coherent (que ca aille toujours dans le
 * meme sens)
 */
struct MONSTER_POS
{
    int xPos;
    int yPos;
    int xPreviousPos;
    int yPreviousPos;
};

/**
 * Nombre de monstres et positions precedentes des monstres
 * les positions des monstres sont utilisees lors du deplacement
 * des monstres, de maniere a ce que leur deplacement soit
 * rectiligne
 */
static int s_nbMonsters;
static MONSTER_POS * s_arrayMonsterPos;

/**
 * Position de depart de pacman
 * Cette position est utilisee lorsque pacman se fait manger
 */
static int s_xFirstPacmanPos;
static int s_yFirstPacmanPos;

/**
 * Position de depart d'un monstre
 * Cette position est utilisee lorsque pacman mange un monstre
 */
static int s_xFirstMonsterPos;
static int s_yFirstMonsterPos;

/**
 * Prochain mouvement relativement a la position actuelle
 * Ces variables sont mises a jour lorsque l'utilisateur 
 * appuie sur les fleches
 */
static int s_xMoveRelative;
static int s_yMoveRelative;

/**
 * Score et etat actuel du jeu
 */
static int s_currentScore;
static int s_lifes;
static int s_levelWon;
static int s_level;

/**
 * Gere le temps pendant lequel pacman peut manger
 * les monstres.
 * Lorsque pacman peut manger des monstres, s_pacmanEatsMonsters
 * vaut 1 et s_timeEnd est l'heure (time_t) jusqu'a quand les 
 * monstres pourront se faire manger.
 * s_timePacEater est le temps relatif en secondes pendant
 * lequel pacman peut manger les monstres
 */
static int s_pacmanEatsMonsters;
static time_t s_timeEnd;
static time_t s_timePacEater;

/**
 * Gere la creation des monstres quand pacman les a manges
 * s_MonsterToCreate indique le nombre de monstres a recreer
 * s_timeRecreateMonster indique l'heure (time_t) a partir de
 * laquelle les monstres pourront etre recrees.
 * s_timeBeginRecreateMonster est le temps relatif en secondes
 * pendant lequel les monstres manges ne sont pas recrees
 */
static int s_MonsterToCreate = 0;
static time_t s_timeRecreateMonster;
static time_t s_timeBeginRecreateMonster;


/**
 * Tableau et taille : Le tableau est alloue a la lecture du fichier
 * s_arrayGame est le tableau de jeu de taille (s_xSize * s_ySize)
 * s_arraySave permet de sauver les fruits lorsque les creatures
 * se deplacent (ceci aurait pu etre optimise)
 */
static int * s_arrayGame;
static int s_xSize;
static int s_ySize;
static int * s_arraySave;

/**
 * Fonctions static non resolues
 */ 
static int is_pacman_current_eater();


/**
 * --statics_init()--
 *
 * Initialisation de toutes les variables static
 */
static void statics_init()
{
    if( s_arrayMonsterPos ) free( s_arrayMonsterPos );
    s_arrayMonsterPos = NULL;
    s_nbMonsters = 0;
    s_xFirstPacmanPos = -1;
    s_yFirstPacmanPos = -1;
    s_xFirstMonsterPos = -1;
    s_yFirstMonsterPos = -1;
    s_xMoveRelative = 0;
    s_yMoveRelative = 0;
    s_currentScore = 0;
    s_lifes = 0;
    s_levelWon = 0;
    s_level = 0;
    s_pacmanEatsMonsters = 0;
	s_timeEnd = 0;
    s_timePacEater = 0;
    s_MonsterToCreate = 0;
    s_timeRecreateMonster = 0;
    s_timeBeginRecreateMonster = 0;
    if( s_arrayGame ) free( s_arrayGame );
    s_arrayGame = NULL;
    s_xSize = 0;
    s_ySize = 0;
    if( s_arraySave ) free( s_arraySave );
    s_arraySave = NULL;
}

/**
 * --pacman_random()
 *
 * @return un nombre entre 0 et maxNumber-1
 */
static int pacman_random( int maxNumber)
{
    // Cette fonction renvoie un nombre entre 0 et maxNumber-1 compris
    // Utilise srand() et rand()
    static int s_first = 1;

    if( s_first )
    {
	s_first = 0;
	srand( (unsigned)time( NULL ) );
    }
    return (rand() % (maxNumber));
}

/**
 * --set_score()--
 *
 * Met a jour le score en fonction du type sur lequel on passe
 */
static void set_score( int type )
{
    switch( type )
    {
    case PAC_TYPE_NOTHING:
        {
            s_currentScore += 1;
            break;
        }
    case PAC_TYPE_MONSTER:
        {
            s_currentScore += 100;
            break;
        }
    case PAC_TYPE_FRUIT:
        {
            s_currentScore += 10;
            break;
        }
    case PAC_TYPE_MAGIC_FRUIT:
        {
            s_currentScore += 50;
            break;
        }
    }
}

/**
 * --update_x_y()--
 *
 * Fait en sorte que 0 <= x < s_xSize
 * et                0 <= y < s_ySize
 *
 * Par exemple si x vaut s_xSize, x est mis a 0
 * Cela gere 'les portes'
 *
 * @param px pointeur sur x
 * @param py pointeur sur y
 */
static void update_x_y( int * px, int * py )
{
    if( *px < 0 ) *px = s_xSize - ((-(*px)) %  s_xSize);
    if( *px >= s_xSize ) *px = *px % s_xSize;

    if( *py < 0 ) *py = s_ySize - ((-(*py)) %  s_ySize);
    if( *py >= s_ySize ) *py = *py % s_ySize;
}

/**
 * --array_get_value()--
 *
 * @param arr le tableau
 * @param x x
 * @param y y
 * @return la valeur a la position x, y
 */
static int array_get_value( int * arr, int x, int y )
{
    update_x_y( &x, &y );
    return arr[(y*s_xSize) + x];
}

/**
 * --array_set_value()--
 *
 * Met la valeur type a la position x, y
 *
 * @param arr le tableau
 * @param x x
 * @param y y
 */
static void array_set_value( int * arr, int x, int y, int type )
{
    update_x_y( &x, &y );
    arr[(y*s_xSize) + x] = type;
}

/**
 * --set_value_to_save()--
 *
 * Met la valeur type a la position x, y
 *
 * @param x x
 * @param y y
 */
static void set_value_to_save( int x, int y, int type )
{
    array_set_value( s_arraySave, x, y, type );
}


/**
 * --get_value_saved()--
 *
 * @param x x
 * @param y y
 * @return la valeur a la position x, y
 */
static int get_value_saved( int x, int y )
{
    return array_get_value( s_arraySave, x, y );
}


/**
 * --set_game_value()--
 *
 * Met la valeur type a la position x, y dans le tableau de jeu
 * et appelle la fonction chargee de l'affichage
 *
 * @param x x
 * @param y y
 */
static void set_game_value( int x, int y, int type )
{
    if( type == PAC_TYPE_PACMAN && (x == 10 || x == 11 || x == 12) && (y == 12 || y == 11 || y == 10 || y == 9 || y == 8 ) )
    {
        x = x;
    }
    update_x_y( &x, &y );
    if( type == PAC_TYPE_PACMAN && (x == 10 || x == 11 || x == 12) && (y == 12 || y == 11 || y == 10 || y == 9 || y == 8 ) )
    {
        x = x;
    }
    display_item( x, y, type, is_pacman_current_eater() );
    array_set_value( s_arrayGame, x, y, type );
}

/**
 * --get_game_value()--
 *
 * @param x x
 * @param y y
 * @return la valeur a la position x, y
 */
static int get_game_value( int x, int y )
{
    return array_get_value( s_arrayGame, x, y );
}



/**
 *  --monster_get_previous_pos()--
 *
 * On recherche la position precedente du monstre, et on met a jour
 * pxMonsterPreviousPos et pyMonsterPreviousPos si on la trouve
 *
 * @param xMonsterPos position actuelle du monstre
 * @param yMonsterPos position actuelle du monstre
 * @param pxMonsterPreviousPos position precedente du monstre en sortie
 * @param pyMonsterPreviousPos position precedente du monstre en sortie
 * @return 0 si trouve, sinon, -1
 */
static int monster_get_previous_pos( int xMonsterPos, int yMonsterPos, int * pxMonsterPreviousPos, int *pyMonsterPreviousPos ) 
{
    for( int i = 0; i < s_nbMonsters; i ++ )
    {
        if( s_arrayMonsterPos[i].xPos == xMonsterPos && s_arrayMonsterPos[i].yPos == yMonsterPos )
        {
            *pxMonsterPreviousPos = s_arrayMonsterPos[i].xPreviousPos;
            *pyMonsterPreviousPos = s_arrayMonsterPos[i].yPreviousPos;
            return 0;
        }
    }
    return -1;
}

/**
 *  --monster_set_new_pos()--
 *
 * On met a jour les nouvelles positions du monstre
 *
 * @param xMonsterNewPos position actuelle du monstre
 * @param yMonsterNewPos position actuelle du monstre
 * @param xMonsterPreviousPos position precedente du monstre
 * @param yMonsterPreviousPos position precedente du monstre
 * @return 0 si trouve, sinon, -1
 */
static int monster_set_new_pos( int xMonsterNewPos, int yMonsterNewPos, int xMonsterPreviousPos, int yMonsterPreviousPos ) 
{
    // (-1,-1) => position particuliere lorsqu'un monstre se fait manger
    if( xMonsterNewPos != -1 && yMonsterNewPos != -1 )
    {
        update_x_y( &xMonsterNewPos, &yMonsterNewPos );
    }
    for( int i = 0; i < s_nbMonsters; i ++ )
    {
        if( s_arrayMonsterPos[i].xPos == xMonsterPreviousPos && s_arrayMonsterPos[i].yPos == yMonsterPreviousPos )
        {
            s_arrayMonsterPos[i].xPreviousPos = s_arrayMonsterPos[i].xPos;
            s_arrayMonsterPos[i].yPreviousPos = s_arrayMonsterPos[i].yPos;
            s_arrayMonsterPos[i].xPos = xMonsterNewPos;
            s_arrayMonsterPos[i].yPos = yMonsterNewPos;
            return 0;
        }
    }
    return -1;
}
/**
 * --monsters_set_pos()--
 *
 * On initialise le tableau s_arrayMonsterPos avec la position de chaque monstre
 */
static void monsters_set_pos() 
{
    int iMonster = 0;
    for( int x = 0; x < s_xSize; x ++ )
    {
        for( int y = 0; y < s_ySize; y ++ )
        {
            int type = get_game_value( x, y );
            if( type == PAC_TYPE_MONSTER && iMonster < s_nbMonsters )
            {
                s_arrayMonsterPos[iMonster].xPreviousPos = x;
                s_arrayMonsterPos[iMonster].yPreviousPos = y;
                s_arrayMonsterPos[iMonster].xPos = x;
                s_arrayMonsterPos[iMonster].yPos = y;
                iMonster ++;
            }
        }
    }
}

/**
 * --get_pacman_pos()--
 *
 * Renvoie la position de pacman
 *
 * @param pxPos la position de pacman en x en sortie
 * @param pyPos la position de pacman en y en sortie
 * @return 0 si OK, -1 si pas trouve
 */
static int get_pacman_pos( int * pxPos, int * pyPos )
{
    for( int x = 0; x < s_xSize; x ++ )
    {
        for( int y = 0; y < s_ySize; y ++ )
        {
            int type = get_game_value( x, y );
            if( type == PAC_TYPE_PACMAN )
            {
                *pxPos = x;
                *pyPos = y;
                return 0;
            }
        }
    }
    // Soit on vient de se faire manger, soit c'est un bug ...
    return -1;
}

/**
 * --move_pacman_to_new_pos()--
 *
 * Met a jour la position de pacman
 *
 * @param xPacPos ancienne position en x
 * @param yPacPos ancienne position en y
 * @param xNewPacPos nouvelle position en x
 * @param yNewPacPos nouvelle position en y
 *
 */
static void move_pacman_to_new_pos( int xPacPos, int yPacPos, int xNewPacPos, int yNewPacPos )
{
    set_game_value( xPacPos, yPacPos, PAC_TYPE_NOTHING );
    set_game_value( xNewPacPos, yNewPacPos, PAC_TYPE_PACMAN );
}



/**
 *  --load_game()--
 *
 * Charge un tableau a partir d'un fichier et initialise
 * toutes les variables.
 * Si un probleme se produit, on quitte le jeu
 *
 * @param gameFile Le nom du fichier
 */
void load_game( const char * gameFile )
{
    int onePacmanPerGame = 0;

    // Initialisation des variables
    statics_init();

    // Charge le tableau et alloue s_arrayGame (!!)
    load_file( gameFile, &s_xSize, &s_ySize, &s_arrayGame );

    // On boucle sur le tableau charge
    for( int x = 0; x < s_xSize; x ++ )
    {
        for( int y = 0; y < s_ySize; y ++ )
        {
            int type = get_game_value( x, y );
            if( type == PAC_TYPE_PACMAN )
            {
                // On fait un petit test s'il y a plus de 1 pacman
                if( onePacmanPerGame == 0 )
                {
                    onePacmanPerGame = 1;
                    
                    // On memorise la position de depart
                    s_xFirstPacmanPos = x;
                    s_yFirstPacmanPos = y;

                    //set_game_value( x, y, type );
                }
                else
                {
                    set_game_value( x, y, PAC_TYPE_NOTHING );
                }
            }
            else
            {
                if( type == PAC_TYPE_MONSTER )
                {
                    // On memorise une position de depart
                    s_xFirstMonsterPos = x;
                    s_yFirstMonsterPos = y;
                    
                    s_nbMonsters ++;
                }
                //set_game_value( x, y, type );
            }
        }
    }

    // Tableau de sauvegarde des fruits sous les creatures
    s_arraySave = (int *)calloc( s_xSize * s_ySize * sizeof(int), 1 );
    if( s_arraySave == NULL )
    {
        error_exit( "Allocation memoire impossible (2) !");
    }    

    // On alloue le tableau de monstres et on l'initialise
    s_arrayMonsterPos = (MONSTER_POS *)calloc( sizeof(MONSTER_POS), s_nbMonsters );
    monsters_set_pos();

    // On rafraichit l'ecran
    repaint();
}

/**
 * --set_level()--
 *
 * Definit le niveau du jeu, avec en particulier, le temps
 * pdt lequel pacman peut manger les monstres apres avoir 
 * mange une fraise, ainsi que le nombre de vies
 *
 * @param level niveau
 * @param timePacEater temps pdt lequel pacman peut manger les monstres
 * @param timeRecreateMonster temps avant que les monstres ne soient recrees
 * @param nbLifes nombre de vies
 */
void set_level( int level, time_t timePacEater, time_t timeRecreateMonster, int nbLifes )
{
    s_level = level;
    s_timePacEater = timePacEater;
    s_timeRecreateMonster = timeRecreateMonster;
    s_lifes = nbLifes;
}



/**
 * --set_current_eater()--
 *
 * Met a jour s_pacmanEatsMonsters si besoin
 */
static void set_current_eater()
{
    time_t t = time(NULL);
    if( t > s_timeEnd )
    {
        s_pacmanEatsMonsters = 0;
    }
    else if( s_timeEnd - t <= 3 )
    {
        s_pacmanEatsMonsters = 2;
    }
    else
    {
        s_pacmanEatsMonsters = 1;
    }
}

/**
 * --set_pacman_current_eater()--
 *
 * Met s_pacmanEatsMonsters a 1, ce qui signifie
 * que pacman peut manger des creatures
 */
static void set_pacman_current_eater()
{
    time_t t = time(NULL);
    s_timeEnd = t + s_timePacEater;
    s_pacmanEatsMonsters = 1;
}

/**
 * --is_pacman_current_eater()--
 *
 * @return Renvoie 1 ou 2 si pacman peut manger les monstres
 */
static int is_pacman_current_eater()
{
    return s_pacmanEatsMonsters;
}



/**
 *  --pacman_eats_a_monster()--
 *
 * @param xMonsterPos Position du monstre mange
 * @param yMonsterPos
 *
 * Met a jour la variable indiquant le nombre de monstres a
 * creer ainsi que le temps a partir duquel on va recreer
 * les monstres
 */
static void pacman_eats_a_monster( int xMonsterPos, int yMonsterPos )
{
    s_MonsterToCreate ++;
    s_timeBeginRecreateMonster = time(NULL) + s_timeRecreateMonster;
    
    // On met a jour les positions du monstre (-1 pour le retrouver)
    monster_set_new_pos( -1, -1, xMonsterPos, yMonsterPos );
}

/**
 *  --recreate_monster_if_needed()--
 *
 * Recree les monstres manges par pacman
 */
static void recreate_monster_if_needed()
{
    // On recree un monstre a la fois, une fois que le temps 
    // s_timeBeginRecreateMonster est ecoule 
    // Le monstre est recree a la position de depart (FirstMonsterPos)
    // a la condition que la case soit vide (cad ni pacman, ni monstre
    // deja la)
    if( s_MonsterToCreate > 0 )
    {
        if( time(NULL) > s_timeBeginRecreateMonster )
        {
            int type = get_game_value( s_xFirstMonsterPos, s_yFirstMonsterPos );
            if( type == PAC_TYPE_NOTHING )
            {
                // Met le monstre ds le tableau de jeu
                set_game_value( s_xFirstMonsterPos, s_yFirstMonsterPos, PAC_TYPE_MONSTER );

                // Met le monstre dans le tableau de position des monstres
                monster_set_new_pos( s_xFirstMonsterPos, s_yFirstMonsterPos, -1, -1 );

                // 1 monstre de moins a creer
                s_MonsterToCreate --;
            }
        }
    }
}

/**
 *  --pacman_loses_a_life()--
 *
 * Decremente le nombre de vie 
 */
static void pacman_loses_a_life()
{
    // On decremente le compteur de vie
    s_lifes --;
}

/**
 *  --pacman_reset_to_initial_position()--
 *
 * Remet pacman a sa position de depart apres avoir eventuellement
 * supprime l monstre qui pourrait se trouver a cette 
 * position
 */
static void pacman_reset_to_initial_position()
{
    if( s_lifes > 0 )
    {
        // On remet le pacman a la position de depart
        // S'il y a un monstre a cette position, alors
        // on l'enleve
        int type = get_game_value( s_xFirstPacmanPos, s_yFirstPacmanPos );
        if( type == PAC_TYPE_MONSTER )
        {
            pacman_eats_a_monster(s_xFirstPacmanPos, s_yFirstPacmanPos);            
        }
        set_game_value( s_xFirstPacmanPos, s_yFirstPacmanPos, PAC_TYPE_PACMAN );
    }
}

/**
 *  --is_monster_can_move()--
 *
 * @param x position du monstre 
 * @param y
 * @return 1 si le monstre peut bouger sinon 0
 */
static int is_monster_can_move( int x, int y )
{
    int type = get_game_value( x, y );
    switch( type )
    {
    case PAC_TYPE_WALL_VERT:
        {
            return 0;
        }
    case PAC_TYPE_WALL_HORIZ:
        {
            return 0;
        }
    case PAC_TYPE_WALL_TOP_LEFT:
        {
            return 0;
        }
    case PAC_TYPE_WALL_TOP_RIGHT:
        {
            return 0;
        }
    case PAC_TYPE_WALL_BOTTOM_LEFT:
        {
            return 0;
        }
    case PAC_TYPE_WALL_BOTTOM_RIGHT:
        {
            return 0;
        }
    //Le monstre a le droit de se suicider
    //case PAC_TYPE_PACMAN:
    //    {
    //        return 0;
    //    }
    case PAC_TYPE_MONSTER:
        {
            return 0;
        }
    }
    return 1;
}

/**
 * --move_monster_to_new_pos()--
 *
 * On deplace le monstre a sa nouvelle position
 *
 * @param xMonsterPos Ancienne position du monstre
 * @param yMonsterPos
 * @param xNewMonsterPos Nouvelle position du monstre
 * @param yNewMonsterPos
 */
static void move_monster_to_new_pos( int xMonsterPos, int yMonsterPos, int xNewMonsterPos, int yNewMonsterPos )
{
    // Si on bouge
    if( xMonsterPos != xNewMonsterPos || yMonsterPos != yNewMonsterPos )
    {
        // On restore ce qu'il y avait sous le monstre
        int savedType = get_value_saved( xMonsterPos, yMonsterPos );
        set_value_to_save( xMonsterPos, yMonsterPos, PAC_TYPE_NOTHING );
        set_game_value( xMonsterPos, yMonsterPos, savedType );

        // On verifie que l'on ne vient pas de se faire manger
        int typeNewPos = get_game_value( xNewMonsterPos, yNewMonsterPos );
        if( typeNewPos == PAC_TYPE_PACMAN )
        {
            // On met a jour le score
            set_score( PAC_TYPE_MONSTER );

            // Qui mange qui ?
            if( is_pacman_current_eater() )
            {
                // Pacman mange le monster et surtout recree un autre monstre
                pacman_eats_a_monster( xMonsterPos, yMonsterPos );                
            }
            else
            {
                // Pacman perd une vie et retourne a la position de depart
                pacman_loses_a_life();                

                // On bouge le monstre a la place de pacman
                set_game_value( xNewMonsterPos, yNewMonsterPos, PAC_TYPE_MONSTER );

                // On met a jour les positions du monstre
                monster_set_new_pos( xNewMonsterPos, yNewMonsterPos, xMonsterPos, yMonsterPos );
            }
        }
        else
        {
            // On sauvegarde la prochaine position
            set_value_to_save( xNewMonsterPos, yNewMonsterPos, typeNewPos );
            
            // On bouge
            set_game_value( xNewMonsterPos, yNewMonsterPos, PAC_TYPE_MONSTER );
            
            // On met a jour les positions du monstre
            monster_set_new_pos( xNewMonsterPos, yNewMonsterPos, xMonsterPos, yMonsterPos );
        }
    }
}


/**
 * --fill_possible_positions()--
 *
 * Remplit le tableau arrPossiblePos avec les positions de deplacement possible
 *
 * @param xMonsterPos position du monstre en x
 * @param yMonsterPos position du monstre en y
 * @param arrRelativesPos tableau des positions relatives
 * @param arrPossiblePos tableau remplit en sortie
 * @param ppossiblePos nombre de positions possibles en sortie
 */
static void fill_possible_positions( int xMonsterPos, int yMonsterPos, int arrRelativesPos[4][2], int arrPossiblePos[4][2], int * ppossiblePos )
{
    // On met toutes les positions possibles dans un tableau
    *ppossiblePos = 0;
    for( int i = 0; i < 4; i ++ )
    {
        int xRel = arrRelativesPos[i][0];
        int yRel = arrRelativesPos[i][1];
        if( is_monster_can_move( xMonsterPos+xRel, yMonsterPos+yRel ) )
        {
            arrPossiblePos[*ppossiblePos][0] = xMonsterPos+xRel;
            arrPossiblePos[*ppossiblePos][1] = yMonsterPos+yRel;
            (*ppossiblePos) ++;
        }
    }
}



/**
 * --init_relative_positions_to_go_closer--
 *
 * On initialise les position relatives pour que le monstre se deplace
 * vers pacman
 *
 * @param xPacPos position en x de pacman
 * @param yPacPos position en y de pacman
 * @param xMonsterPos position en x du monstre
 * @param yMonsterPos position en y du monstre
 * @param arrRelativesPos tableau a initialiser
 */
void init_relative_positions_to_go_closer( int xPacPos, int yPacPos, int xMonsterPos, int yMonsterPos, int arrRelativesPos[4][2] )
{
    int iPos = 0;
    int xDiff = xPacPos - xMonsterPos;
    if( xDiff < 0 ) xDiff = -xDiff;
    int yDiff = yPacPos - yMonsterPos;
    if( yDiff < 0 ) yDiff = -yDiff;

    if( xDiff < yDiff && xDiff > 0 || yDiff == 0 )
    {
        // On commence par les x
        arrRelativesPos[0][0] = (xPacPos - xMonsterPos > 0 ? 1 : -1 );
        arrRelativesPos[0][1] = 0;
        arrRelativesPos[1][0] = 0;
        arrRelativesPos[1][1] = (yPacPos - yMonsterPos > 0 ? 1 : -1 );
        arrRelativesPos[2][0] = (xPacPos - xMonsterPos > 0 ? -1 : 1 );
        arrRelativesPos[2][1] = 0;
        arrRelativesPos[3][0] = 0;
        arrRelativesPos[3][1] = (yPacPos - yMonsterPos > 0 ? -1 : 1 );
    }
    else
    {
        // On commence par les y
        arrRelativesPos[0][0] = 0;
        arrRelativesPos[0][1] = (yPacPos - yMonsterPos > 0 ? 1 : -1 );
        arrRelativesPos[1][0] = (xPacPos - xMonsterPos > 0 ? 1 : -1 );
        arrRelativesPos[1][1] = 0;
        arrRelativesPos[2][0] = 0;
        arrRelativesPos[2][1] = (yPacPos - yMonsterPos > 0 ? -1 : 1 );
        arrRelativesPos[3][0] = (xPacPos - xMonsterPos > 0 ? -1 : 1 );
        arrRelativesPos[3][1] = 0;
    }
}

/**
 * --init_relative_positions_to_go_further--
 *
 * On initialise les position relatives pour que le monstre se deplace
 * plus loin de pacman
 *
 * @param xPacPos position en x de pacman
 * @param yPacPos position en y de pacman
 * @param xMonsterPos position en x du monstre
 * @param yMonsterPos position en y du monstre
 * @param arrRelativesPos tableau a initialiser
 */
void init_relative_positions_to_go_further( int xPacPos, int yPacPos, int xMonsterPos, int yMonsterPos, int arrRelativesPos[4][2] )
{
    // On appelle init_relative_positions_to_go_closer()
    init_relative_positions_to_go_closer( xPacPos, yPacPos, xMonsterPos, yMonsterPos, arrRelativesPos );

    // On change les signes
    for( int i = 0; i < 4 ; i ++ )
    {
        arrRelativesPos[i][0] = arrRelativesPos[i][0] * (-1);
        arrRelativesPos[i][1] = arrRelativesPos[i][1] * (-1);
    }
}

/**
 * --monster_flees_pacman()--
 *
 * le monstre fuit pacman
 *
 * @param xPacPos position en x de pacman
 * @param yPacPos position en y de pacman
 * @param xMonsterPos position en x du monstre
 * @param yMonsterPos position en y du monstre
 */
static void monster_flees_pacman( int xPacPos, int yPacPos, int xMonsterPos, int yMonsterPos )
{
    int arrPossiblePos[4][2];
    int possiblePos = 0;
    int xNewMonsterPos = xMonsterPos;
    int yNewMonsterPos = yMonsterPos;
    int arrRelativesPos[4][2] = { {0,0}, {0,0}, {0,0}, {0,0} };

    // On initialise les positions relatives
    init_relative_positions_to_go_further( xPacPos, yPacPos, xMonsterPos, yMonsterPos, arrRelativesPos );

    // On met toutes les positions possibles dans un tableau
    fill_possible_positions( xMonsterPos, yMonsterPos, arrRelativesPos, arrPossiblePos, &possiblePos );

    // On prend la premiere position
    if( possiblePos > 0 )
    {
        xNewMonsterPos = arrPossiblePos[0][0];
        yNewMonsterPos = arrPossiblePos[0][1];
    }

    // On se deplace ici
    move_monster_to_new_pos( xMonsterPos, yMonsterPos, xNewMonsterPos, yNewMonsterPos );
}

/**
 * --monster_flees_pacman()--
 *
 * le monstre se rapproche de pacman
 *
 * @param xPacPos position en x de pacman
 * @param yPacPos position en y de pacman
 * @param xMonsterPos position en x du monstre
 * @param yMonsterPos position en y du monstre
 */
static void monster_goes_closer_to_pacman( int xPacPos, int yPacPos, int xMonsterPos, int yMonsterPos )
{
    int arrPossiblePos[4][2];
    int possiblePos = 0;
    int xNewMonsterPos = xMonsterPos;
    int yNewMonsterPos = yMonsterPos;
    int arrRelativesPos[4][2] = { {0,0}, {0,0}, {0,0}, {0,0} };

    // On initialise les positions relatives
    init_relative_positions_to_go_closer( xPacPos, yPacPos, xMonsterPos, yMonsterPos, arrRelativesPos );

    // On met toutes les positions possibles dans un tableau
    fill_possible_positions( xMonsterPos, yMonsterPos, arrRelativesPos, arrPossiblePos, &possiblePos );

    // On prend la premiere position
    if( possiblePos > 0 )
    {
        xNewMonsterPos = arrPossiblePos[0][0];
        yNewMonsterPos = arrPossiblePos[0][1];
    }

    // On se deplace ici
    move_monster_to_new_pos( xMonsterPos, yMonsterPos, xNewMonsterPos, yNewMonsterPos );
}

/**
 *  --monster_goes_randomly()
 *
 * Le monstre se deplace au hasard
 *
 * @param xMonsterPos position en x du monstre
 * @param yMonsterPos position en y du monstre
 */
static void monster_goes_randomly( int xMonsterPos, int yMonsterPos )
{
    int arrPossiblePos[4][2];
    int possiblePos = 0;
    int arrRelativesPos[4][2] = { {1,0}, {-1,0}, {0,1}, {0,-1} };
    int xNewMonsterPos = xMonsterPos;
    int yNewMonsterPos = yMonsterPos;

    // On met toutes les positions possibles dans un tableau
    fill_possible_positions( xMonsterPos, yMonsterPos, arrRelativesPos, arrPossiblePos, &possiblePos );

    // On prend une des positions au hasard
    if( possiblePos > 0 )
    {
        int nRandomPos = pacman_random( possiblePos );
    
        // D'ou la nouvelle position
        xNewMonsterPos = arrPossiblePos[nRandomPos][0];
        yNewMonsterPos = arrPossiblePos[nRandomPos][1];
    }

    // On se deplace ici
    move_monster_to_new_pos( xMonsterPos, yMonsterPos, xNewMonsterPos, yNewMonsterPos );
}



/**
 *  --monster_goes_linear()
 *
 * Le monstre se deplace dans la meme direction que precedement
 *
 * @param xMonsterPos position en x du monstre
 * @param yMonsterPos position en y du monstre
 */
static void monster_goes_linear( int xMonsterPos, int yMonsterPos )
{
    int xMonsterPreviousPos, yMonsterPreviousPos;
    
    // On recherche la position precedente
    if( monster_get_previous_pos( xMonsterPos, yMonsterPos, &xMonsterPreviousPos, &yMonsterPreviousPos ) == -1 )
    {
        // position non trouvee : est-ce normal
        printf( "kk\n" );
        // monster_goes_randomly( xMonsterPos, yMonsterPos );
    }
    else
    {
        // On calcule le deplacement relatif
        int xRel = xMonsterPos-xMonsterPreviousPos;
        int yRel = yMonsterPos-yMonsterPreviousPos;

        // Peut-on bouger ici ?
        if( ! is_monster_can_move( xMonsterPos+xRel, yMonsterPos+yRel ) )
        {
            // Non, on se deplace au hasard
            monster_goes_randomly( xMonsterPos, yMonsterPos );
        }
        else
        {
            // Oui, on se deplace a cette position
            move_monster_to_new_pos( xMonsterPos, yMonsterPos, xMonsterPos+xRel, yMonsterPos+yRel );
        }
    }

}


/**
 * --move_monster()--
 *
 * Deplace un monstre
 *
 * @param xPacPos position en x de pacman
 * @param yPacPos position en y de pacman
 * @param xMonsterPos position en x du monstre
 * @param yMonsterPos position en y du monstre
 */
static void move_monster( int xPacPos, int yPacPos, int xMonsterPos, int yMonsterPos )
{
    int xDiff = xPacPos - xMonsterPos;
    if( xDiff < 0 ) xDiff = -xDiff;
    int yDiff = yPacPos - yMonsterPos;
    if( yDiff < 0 ) yDiff = -yDiff;

    if( (xDiff + yDiff) < ((s_xSize / 4) + (s_ySize / 4)) )
    {
        // On regarde si on doit craindre pacman
        if( is_pacman_current_eater() )
        {
            // On s'eloigne de pacman
            monster_flees_pacman( xPacPos, yPacPos, xMonsterPos, yMonsterPos );
        }
        else
        {
            // On se rapproche de pacman
            monster_goes_closer_to_pacman( xPacPos, yPacPos, xMonsterPos, yMonsterPos );
        }
    }
    else
    {
        // On se deplace lineairement
        monster_goes_linear( xMonsterPos, yMonsterPos );
    }
}

/**
 * --level_state()--
 *
 * Regarde s'il reste des trucs a manger ds le tableau
 */
static void level_state()
{
    // Puis chaque valeur
    for( int x = 0; x < get_x_size(); x ++ )
    {
        for( int y = 0; y < get_y_size(); y ++ )
        {
            int type = get_game_value( x, y );
            switch( type )
            {
            case PAC_TYPE_MAGIC_FRUIT:
            case PAC_TYPE_FRUIT:
                {
                    // Tableau pas vide
                    return;
                }
            case PAC_TYPE_MONSTER:
                {
                    int typeSaved = get_value_saved( x, y );
                    if( typeSaved == PAC_TYPE_MAGIC_FRUIT || typeSaved == PAC_TYPE_FRUIT )
                    {
                        // Tableau pas vide
                        return;
                    }
                    break;
                }
            }
    
        }
    }
    s_levelWon = 1;
}


/**
 * --move_monsters()--
 *
 * Deplace les monstres
 */
void move_monsters()
{
    int xPacPos, yPacPos;
    
    // Met a jour qui mange qui
    set_current_eater();

    // Recree un monstre si necessaire
    recreate_monster_if_needed();

    // Recupere la position de pacman
    if( get_pacman_pos( &xPacPos, &yPacPos ) == 0 )
    {
        // On boucle sur le tableau de monstres 
        for( int i = 0; i < s_nbMonsters; i ++ )
        {
            move_monster( xPacPos, yPacPos, 
                s_arrayMonsterPos[i].xPos, s_arrayMonsterPos[i].yPos );
        }
    }

    // On remet pacman au depart s'il s'est fait mange
    if( get_pacman_pos( &xPacPos, &yPacPos ) < 0 )
    {
        pacman_reset_to_initial_position();
    }

    // Affiche les infos
    display_infos();

    // Regarde s'il reste des trucs ds le tableau
    level_state();
}

/**
 * --move_pacman()--
 * 
 * Deplace pacman
 */
void move_pacman()
{
    int xPacPos, yPacPos, type;
    
    // Met a jour qui mange qui
    set_current_eater();

    // Recree un monstre si necessaire
    recreate_monster_if_needed();

    // Recupere la position de pacman
    if( get_pacman_pos( &xPacPos, &yPacPos ) == 0 )
    {
        // Recupere le type
        type = get_game_value( xPacPos + s_xMoveRelative, yPacPos + s_yMoveRelative );
        
        // Met a jour le score en fonction du type
        set_score( type );

        // Determine l'action (deplacement...)
        switch( type )
        {
        case PAC_TYPE_NOTHING:
        case PAC_TYPE_FRUIT:
            {
                // On se deplace a la nouvelle position
                move_pacman_to_new_pos( xPacPos, yPacPos, xPacPos + s_xMoveRelative, yPacPos + s_yMoveRelative );
                break;
            }
        case PAC_TYPE_MONSTER:
            {
                // Qui mange qui ?
                if( is_pacman_current_eater() )
                {
                    // Mange le monster et surtout recree un autre monstre
                    pacman_eats_a_monster(xPacPos + s_xMoveRelative, yPacPos + s_yMoveRelative);
                    move_pacman_to_new_pos( xPacPos, yPacPos, xPacPos + s_xMoveRelative, yPacPos + s_yMoveRelative );
                }
                else
                {
                    // Perd une vie et retourne a la position de depart
                    pacman_loses_a_life();
                    
                    // On remplace la case par rien
                    set_game_value( xPacPos, yPacPos, PAC_TYPE_NOTHING );
                }
                break;
            }
        case PAC_TYPE_MAGIC_FRUIT:
            {
                // Le pacman peut manger les creatures
                set_pacman_current_eater();
                
                // On se deplace a la nouvelle position
                move_pacman_to_new_pos( xPacPos, yPacPos, xPacPos + s_xMoveRelative, yPacPos + s_yMoveRelative );
                break;
            }
        }
    }

    // On remet pacman au depart s'il s'est fait mange
    if( get_pacman_pos( &xPacPos, &yPacPos ) < 0 )
    {
        pacman_reset_to_initial_position();
    }

    // Affiche les infos
    display_infos();

    // Regarde s'il reste des trucs ds le tableau
    level_state();
}

/**
 * --go_left()--
 *
 * Demande a aller a gauche
 */
void go_left()
{
    s_xMoveRelative = -1;
    s_yMoveRelative = 0;
}

/**
 * --go_right()--
 *
 * Demande a aller a droite
 */
void go_right()
{
    s_xMoveRelative = 1;
    s_yMoveRelative = 0;
}

/**
 * --go_top()--
 *
 * Demande a aller vers le haut
 */
void go_top()
{
    s_xMoveRelative = 0;
    s_yMoveRelative = -1;
}

/**
 * --go_bottom()--
 *
 * Demande a aller vers le bas
 */
void go_bottom()
{
    s_xMoveRelative = 0;
    s_yMoveRelative = 1;
}

/**
 * --is_game_over()--
 *
 * Renvoie 1 si la partie est perdue
 */
int is_game_over()
{
    return s_lifes <= 0;
}

/**
 * --is_game_won()--
 *
 * Renvoie 1 si le tableau est gane
 */
int is_level_won()
{
    return s_levelWon;
}

/**
 * --get_score()--
 *
 * Renvoie le score actuel
 */
int get_score()
{
    return s_currentScore;
}

/**
 * --get_level()--
 *
 * Renvoie le niveau
 */
int get_level()
{
    return s_level;
}

/**
 * --get_lifes()--
 *
 * Renvoie le nombre de vies
 */
int get_lifes()
{
    return s_lifes;
}


/**
 * --get_x_size()--
 *
 * Renvoie la taille du tableau en x
 */
int get_x_size()
{
    return s_xSize;
}

/**
 * --get_y_size()--
 *
 * Renvoie la taille du tableau en y
 */
int get_y_size()
{
    return s_ySize;
}

/**
 * --repaint()--
 *
 * Redessine tout l'ecran
 */
void repaint()
{
    // dessine deja le fond
    display_screen();
    
    // Puis les infos
    display_infos();

    // Puis chaque valeur
    for( int x = 0; x < get_x_size(); x ++ )
    {
        for( int y = 0; y < get_y_size(); y ++ )
        {
            int type = get_game_value( x, y );
            int pacEater = is_pacman_current_eater();
            display_item( x, y, type, pacEater );
        }
    }
}
