/**
 * @file
 *
 * Ce fichier permet de creer des tableaux
 *
 * @author Sylvain MARECHAL
 * @date $Id$
 */
#ifdef _WIN32
#pragma warning(disable:4996) /*Suppression du warning ""'' was declared deprecated""*/
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "pac_core.h"
#include "tab_core.h"
#include "tab_view.h"
#include "pac_err.h"
#include "pac_file.h"

/**
 * Tableau et taille : Le tableau est alloue a la lecture du fichier
 * s_arrayGame est le tableau de jeu de taille (s_xSize * s_ySize)
 */
static int * s_arrayGame;
static int s_xSize;
static int s_ySize;

/**
 * Position du curseur
 */
static int s_x;
static int s_y;

/**
 * Position precedente et type
 */
static int s_xLast;
static int s_yLast;
static int s_lastType;

/**
 * --statics_init()--
 *
 * Initialisation de toutes les variables static
 */
static void statics_init()
{
    if( s_arrayGame ) free( s_arrayGame );
    s_arrayGame = NULL;
    s_xSize = 0;
    s_ySize = 0;
    s_x = 0;
    s_y = 0;
    s_xLast = 0;
    s_yLast = 0;
    s_lastType = PAC_TYPE_NOTHING;
}

/**
 * --tab_get_x_size()--
 *
 * Renvoie la taille du tableau en x
 */
int tab_get_x_size()
{
    return s_xSize;
}

/**
 * --tab_get_y_size()--
 *
 * Renvoie la taille du tableau en y
 */
int tab_get_y_size()
{
    return s_ySize;
}

/**
 * --update_x_y()--
 *
 * Fait en sorte que 0 <= x < s_xSize
 * et                0 <= y < s_ySize
 *
 * Par exemple si x vaut s_xSize, x est mis a 0
 * Cela gere 'les portes'
 *
 * @param px pointeur sur x
 * @param py pointeur sur y
 */
static void update_x_y( int * px, int * py )
{
    if( *px < 0 ) *px = s_xSize - ((-(*px)) %  s_xSize);
    if( *px >= s_xSize ) *px = *px % s_xSize;

    if( *py < 0 ) *py = s_ySize - ((-(*py)) %  s_ySize);
    if( *py >= s_ySize ) *py = *py % s_ySize;
}

/**
 * --tab_array_get_value()--
 *
 * @param x x
 * @param y y
 * @return la valeur a la position x, y
 */
static int tab_array_get_value( int x, int y )
{
    update_x_y( &x, &y );
    return s_arrayGame[(y*s_xSize) + x];
}

/**
 * --tab_array_set_value()--
 *
 * Met la valeur type a la position x, y
 *
 * @param x x
 * @param y y
 */
static void tab_array_set_value( int x, int y, int type )
{
    update_x_y( &x, &y );
    s_arrayGame[(y*s_xSize) + x] = type;
}



/**
 * --tab_new()--
 *
 * Cree un nouveau tableau et rafrachit l'ecran
 *
 * @param x taille en x du tableau
 * @param y taille en y du tableau
 */
void tab_new( int x, int y )
{
    // Init statics
    statics_init();

    // Taille du tableau en cases
    s_xSize = x;
    s_ySize = y;

    // Allocation du tableau de jeu
    s_arrayGame = (int *)calloc( s_xSize * s_ySize * sizeof(int), 1 );
    if( s_arrayGame == NULL )
    {
        error_exit( "Allocation memoire impossible (tab)!");
    }

    // Affiche ecran
    tab_repaint();
}

/**
 * --tab_open()--
 *
 * Ouvre un tableau en vue de le modifier
 * et rafrachit l'ecran
 *
 * @param pszFile Fichier contenant les donnees du tableau
 */
void tab_open( const char * pszFile )
{
    // Init statics
    statics_init();

    // Charge le tableau
    load_file( pszFile, 
        &s_xSize, &s_ySize, &s_arrayGame );

    // Affiche ecran
    tab_repaint();
}



/**
 * --tab_save()--
 *
 * Sauvegarde un tableau 
 *
 * @param pszFile Fichier 
 */
void tab_save( const char * pszFile )
{
    save_file( pszFile, s_xSize, s_ySize, s_arrayGame );
}

/**
 * --tab_set_and_refresh()--
 *
 * Met la piece a la position actuelle
 * puis rafraichit l'affichage
 */
void tab_set_and_refresh( int type )
{
    update_x_y( &s_x, &s_y );
    tab_array_set_value( s_x, s_y, type );
    tab_display_item( s_x, s_y, type );
}

/**
 * --tab_wall_vert()--
 *
 * Met la piece a la position actuelle
 * puis rafraichit l'affichage
 */
void tab_wall_vert()
{
    tab_set_and_refresh( PAC_TYPE_WALL_VERT );
    tab_bottom();
}

/**
 * --tab_wall_horiz()--
 *
 * Met la piece a la position actuelle
 * puis rafraichit l'affichage
 */
void tab_wall_horiz()
{
    tab_set_and_refresh( PAC_TYPE_WALL_HORIZ );
    tab_right();
}

/**
 * --tab_wall_top_left()--
 *
 * Met la piece a la position actuelle
 * puis rafraichit l'affichage
 */
void tab_wall_top_left()
{
    tab_set_and_refresh( PAC_TYPE_WALL_TOP_LEFT );
    tab_right();
}

/**
 * --tab_wall_top_right()--
 *
 * Met la piece a la position actuelle
 * puis rafraichit l'affichage
 */
void tab_wall_top_right()
{
    tab_set_and_refresh( PAC_TYPE_WALL_TOP_RIGHT );
    tab_bottom();
}

/**
 * --tab_wall_bottom_left()--
 *
 * Met la piece a la position actuelle
 * puis rafraichit l'affichage
 */
void tab_wall_bottom_left()
{
    tab_set_and_refresh( PAC_TYPE_WALL_BOTTOM_LEFT );
    tab_right();
}

/**
 * --tab_wall_bottom_right()--
 *
 * Met la piece a la position actuelle
 * puis rafraichit l'affichage
 */
void tab_wall_bottom_right()
{
    tab_set_and_refresh( PAC_TYPE_WALL_BOTTOM_RIGHT );
}

/**
 * --tab_fruit()--
 *
 * Met la piece a la position actuelle
 * puis rafraichit l'affichage
 */
void tab_fruit()
{
    tab_set_and_refresh( PAC_TYPE_FRUIT );
}

/**
 * --tab_magic_fruit()--
 *
 * Met la piece a la position actuelle
 * puis rafraichit l'affichage
 */
void tab_magic_fruit()
{
    tab_set_and_refresh( PAC_TYPE_MAGIC_FRUIT );
}

/**
 * --tab_pacman()--
 *
 * Met la piece a la position actuelle
 * puis rafraichit l'affichage
 */
void tab_pacman()
{
    tab_set_and_refresh( PAC_TYPE_PACMAN );
}

/**
 * --tab_monster()--
 *
 * Met la piece a la position actuelle
 * puis rafraichit l'affichage
 */
void tab_monster()
{
    tab_set_and_refresh( PAC_TYPE_MONSTER );
}

void tab_nothing()
{
    tab_set_and_refresh( PAC_TYPE_NOTHING );
}

/**
 * --tab_repeat()--
 *
 * Repete la derniere action
 */
void tab_repeat()
{
}

/**
 * --tab_left()--
 *
 * Met la piece a la position actuelle
 * puis rafraichit l'affichage
 */
void tab_left()
{
    s_xLast = s_x;
    s_x --;
    int type = tab_array_get_value( s_x, s_y );
    tab_set_and_refresh( type );
}

/**
 * --tab_right()--
 *
 * Met la piece a la position actuelle
 * puis rafraichit l'affichage
 */
void tab_right()
{
    s_xLast = s_x;
    s_x ++;
    int type = tab_array_get_value( s_x, s_y );
    tab_set_and_refresh( type );
}

/**
 * --tab_top()--
 *
 * Met la piece a la position actuelle
 * puis rafraichit l'affichage
 */
void tab_top()
{
    s_yLast = s_y;
    s_y --;
    int type = tab_array_get_value( s_x, s_y );
    tab_set_and_refresh( type );
}

/**
 * --tab_bottom()--
 *
 * Met la piece a la position actuelle
 * puis rafraichit l'affichage
 */
void tab_bottom()
{
    s_yLast = s_y;
    s_y ++;
    int type = tab_array_get_value( s_x, s_y );
    tab_set_and_refresh( type );
}

/**
 * --tab_repaint()--
 *
 */
void tab_repaint()
{
    tab_display_screen();
    for( int x = 0; x < tab_get_x_size(); x ++ )
    {
        for( int y = 0; y < tab_get_y_size(); y ++ )
        {
            tab_display_item( x, y, tab_array_get_value( x, y ) );
        }
    }
}